import type { Logger } from 'winston';
import net from 'net';
import type { Socket } from 'net';
import { createLog } from '../log'
import { Shell } from '../shell';
import { delay } from '../delay';

enum Action {
	NONE,
	LOAD,
	DISPLAY,
	STOP
}

enum Mode {
	RGB,
	BW,
	INVERT,
	BW_INVERT,
	RGB_CHANNELS,
	INVERT_CHANNELS
}

export interface fdOutgoingMessage {
	action : Action,
	image : string,
	mode? : Mode,
	x? : number,
	y? : number,
	w? : number,
	h? : number,
	exposure? : number[]
}

export interface fdIncomingMessage {
	action : Action,
	success : boolean
}

export class FD {
	private bin : string;
	private width : number;
	private height : number;
	private host : string;
	private port : number;
	private shell : Shell;
	private log : Logger;
	private client : Socket;
	private socketAvailable : boolean = false;
	private socketConnected : boolean = false;

	constructor (bin: string, width : number, height : number, host : string, port : number) {
		this.bin = bin;
		this.width = width;
		this.height = height;
		this.host = host;
		this.port = port;
		this.log = createLog('fd');
		this.shell = new Shell([ this.bin, `${this.width}`, `${this.height}`, `${this.port}` ], this.logstd.bind(this), this.logsterr.bind(this), null, true);
		this.startDisplay();
		this.startClient();
		this.test();
	}

	private async startDisplay () {
		this.log.info(`Launching fd binary ${this.bin}`);
		this.shell.execute();
	}

	private async startClient () {
		this.client = new net.Socket();
		this.log.info(`Waiting for TCP socket server on ${this.host}:${this.port}...`);
		while (!this.socketAvailable) {
			await delay(10);
		}
		this.log.info(`TCP socket server on ${this.host}:${this.port} is available`);
		this.client.connect(this.port, this.host, async function () {
			this.socketConnected = true;
			this.log.info(`TCP socket client connected to ${this.host}:${this.port}`);
			await delay(1000);
			const msg : fdOutgoingMessage = {
				action : Action.LOAD,
				image : './test.jpg'
			};
			this.send(msg);
		}.bind(this));

		this.client.on('data', function (data : string) {
		    this.receive(data);
		}.bind(this));

		this.client.on('close', () => {
		    this.log.info('Closing connection');
		});

		this.client.on('error', (err : Error) => {
		    this.log.error('Error:', err);
		});
	}

	private logstd (data : string ) {
		this.log.info(`[shell] ${data}`);
		if (data.indexOf('TCP server listening on port') !== -1 && data.indexOf(`${this.port}`) !== -1) {
			this.socketAvailable = true;
		}
	}

	private logsterr (data : string) {
		this.log.error(`[shell] ${data}`);
	}

	private async test () {
		await delay(10000);
		this.shell.kill();
	}

	private send (msg : fdOutgoingMessage) {
		const json : string = JSON.stringify(msg);
		this.log.info(json);
		this.client.write(json);
	}

	private receive (json : string) : fdIncomingMessage {
		const msg : fdIncomingMessage = JSON.parse(json);
		this.log.info(msg);
		return msg;
	}

	public async load (image : string, x : number, y : number, w : number, h : number) : Promise<boolean> {
		const msg : fdOutgoingMessage = {
			action : Action.LOAD,
			image,
			x,
			y,
			w,
			h
		};
		this.send(msg);
		return true;
	}

	public async display (image : string, exposure : number[]) : Promise<boolean> {
		const msg : fdOutgoingMessage = {
			action : Action.DISPLAY,
			image,
			exposure
		};
		this.send(msg);
		return true;
	}

	public async stop (image : string) : Promise<boolean> {
		const msg : fdOutgoingMessage = {
			action : Action.STOP,
			image
		};
		this.send(msg);
		return true;
	}
}

